@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2007 OpenedHand Ltd.
+ * Copyright (C) 2007, 2008 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
@@ -32,13 +32,12 @@
 #include <libsoup/soup-date.h>
 #include <uuid/uuid.h>
 #include <string.h>
-
 #include "gupnp-service.h"
-#include "gupnp-service-private.h"
+#include "gupnp-root-device.h"
 #include "gupnp-context-private.h"
 #include "gupnp-marshal.h"
 #include "gupnp-error.h"
-#include "accept-language.h"
+#include "http-headers.h"
 #include "gena-protocol.h"
 #include "xml-util.h"
 #include "gvalue-util.h"
@@ -59,9 +58,6 @@ struct _GUPnPServicePrivate {
         GQueue          *notify_queue;
 
         gboolean         notify_frozen;
-
-        GList           *pending_messages; /* Pending SoupMessages from this
-                                              service */
 };
 
 enum {
@@ -78,7 +74,7 @@ enum {
 
 static guint signals[LAST_SIGNAL];
 
-static xmlChar *
+static char *
 create_property_set (GQueue *queue);
 
 static void
@@ -95,11 +91,37 @@ typedef struct {
         int           seq;
 
         guint         timeout_id;
+
+        GList        *pending_messages; /* Pending SoupMessages from this
+                                           subscription */
 } SubscriptionData;
 
 static void
 subscription_data_free (SubscriptionData *data)
 {
+        GUPnPContext *context;
+        SoupSession *session;
+
+        context = gupnp_service_info_get_context
+                        (GUPNP_SERVICE_INFO (data->service));
+        session = _gupnp_context_get_session (context);
+
+        /* Cancel pending messages */
+        while (data->pending_messages) {
+                SoupMessage *msg;
+
+                msg = data->pending_messages->data;
+
+                soup_session_cancel_message (session,
+                                             msg,
+                                             SOUP_STATUS_CANCELLED);
+
+                data->pending_messages =
+                        g_list_delete_link (data->pending_messages,
+                                            data->pending_messages);
+        }
+       
+        /* Further cleanup */
         while (data->callbacks) {
                 g_free (data->callbacks->data);
                 data->callbacks = g_list_delete_link (data->callbacks,
@@ -135,13 +157,15 @@ struct _GUPnPServiceAction {
 
         xmlNode     *node;
 
-        xmlNode     *response_node;
+        GString     *response_str;
 };
 
 /**
  * gupnp_service_action_get_name
  * @action: A #GUPnPServiceAction
  *
+ * Get the name of @action.
+ *
  * Return value: The name of @action
  **/
 const char *
@@ -156,8 +180,10 @@ gupnp_service_action_get_name (GUPnPServiceAction *action)
  * gupnp_service_action_get_locales
  * @action: A #GUPnPServiceAction
  *
- * Return value: An ordered (preferred first) #GList of locales preferred by
+ * Get an ordered (preferred first) #GList of locales preferred by
  * the client. Free list and elements after use.
+ *
+ * Return value: A #GList of #char* locale names.
  **/
 GList *
 gupnp_service_action_get_locales (GUPnPServiceAction *action)
@@ -246,19 +272,26 @@ gupnp_service_action_get_value (GUPnPServiceAction *action,
                                 GValue             *value)
 {
         xmlNode *node;
+        gboolean found;
 
         g_return_if_fail (action != NULL);
         g_return_if_fail (argument != NULL);
         g_return_if_fail (value != NULL);
 
+        found = FALSE;
         for (node = action->node->children; node; node = node->next) {
                 if (strcmp ((char *) node->name, argument) != 0)
                         continue;
 
-                xml_util_node_get_content_value (node, value);
+                found = gvalue_util_set_value_from_xml_node (value, node);
 
                 break;
         }
+
+        if (!found)
+                g_warning ("Failed to retreive '%s' argument of '%s' action",
+                           argument,
+                           action->name);
 }
 
 /**
@@ -340,13 +373,11 @@ gupnp_service_action_set_value (GUPnPServiceAction *action,
                                 const char         *argument,
                                 const GValue       *value)
 {
-        char *str;
-
         g_return_if_fail (action != NULL);
         g_return_if_fail (argument != NULL);
         g_return_if_fail (value != NULL);
 
-        if (action->msg->status == SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+        if (action->msg->status_code == SOUP_STATUS_INTERNAL_SERVER_ERROR) {
                 g_warning ("Calling gupnp_service_action_set_value() after "
                            "having called gupnp_service_action_return_error() "
                            "is not allowed.");
@@ -354,17 +385,10 @@ gupnp_service_action_set_value (GUPnPServiceAction *action,
                 return;
         }
 
-        /* Transform to string */
-        str = gvalue_util_value_get_string (value);
-        if (str != NULL) {
-                /* Append to response */
-                xmlNewTextChild (action->response_node,
-                                 NULL,
-                                 (const xmlChar *) argument,
-                                 (xmlChar *) str);
-
-                g_free (str);
-        }
+        /* Append to response */
+        xml_util_start_element (action->response_str, argument);
+        gvalue_util_value_append_to_xml_string (value, action->response_str);
+        xml_util_end_element (action->response_str, argument);
 }
 
 /**
@@ -398,10 +422,6 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
                                    guint               error_code,
                                    const char         *error_description)
 {
-        xmlNode *node;
-        xmlNs *ns;
-        char *tmp;
-
         g_return_if_fail (action != NULL);
 
         switch (error_code) {
@@ -430,47 +450,37 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
                 break;
         }
 
-        /* Replace response_node with a SOAP Fault */
-        xmlFreeNode (action->response_node);
-        action->response_node = xmlNewNode (NULL,
-                                            (const xmlChar *) "Fault");
+        /* Replace response_str with a SOAP Fault */
+        g_string_erase (action->response_str, 0, -1);
 
-        xmlNewTextChild (action->response_node,
-                         NULL,
-                         (const xmlChar *) "faultcode",
-                         (const xmlChar *) "s:Client");
+        xml_util_start_element (action->response_str, "s:Fault");
 
-        xmlNewTextChild (action->response_node,
-                         NULL,
-                         (const xmlChar *) "faultstring",
-                         (const xmlChar *) "UPnPError");
+        xml_util_start_element (action->response_str, "faultcode");
+        g_string_append (action->response_str, "s:Client");
+        xml_util_end_element (action->response_str, "faultcode");
 
-        node = xmlNewTextChild (action->response_node,
-                                NULL,
-                                (const xmlChar *) "detail",
-                                NULL);
+        xml_util_start_element (action->response_str, "faultstring");
+        g_string_append (action->response_str, "UPnPError");
+        xml_util_end_element (action->response_str, "faultstring");
 
-        node = xmlNewTextChild (node,
-                                NULL,
-                                (const xmlChar *) "UPnPError",
-                                NULL);
+        xml_util_start_element (action->response_str, "detail");
 
-        ns = xmlNewNs (node,
-                       (const xmlChar *) "urn:schemas-upnp-org:control-1-0",
-                       (const xmlChar *) "u");
-        xmlSetNs (node, ns);
+        xml_util_start_element (action->response_str,
+                                "UPnPError "
+                                "xmlns=\"urn:schemas-upnp-org:control-1-0\"");
 
-        tmp = g_strdup_printf ("%u", error_code);
-        xmlNewTextChild (node,
-                         NULL,
-                         (const xmlChar *) "errorCode",
-                         (const xmlChar *) tmp);
-        g_free (tmp);
+        xml_util_start_element (action->response_str, "errorCode");
+        g_string_append_printf (action->response_str, "%u", error_code);
+        xml_util_end_element (action->response_str, "errorCode");
 
-        xmlNewTextChild (node,
-                         NULL,
-                         (const xmlChar *) "errorDescription",
-                         (const xmlChar *) error_description);
+        xml_util_start_element (action->response_str, "errorDescription");
+        xml_util_add_content (action->response_str, error_description);
+        xml_util_end_element (action->response_str, "errorDescription");
+
+        xml_util_end_element (action->response_str, "UPnPError");
+        xml_util_end_element (action->response_str, "detail");
+
+        xml_util_end_element (action->response_str, "s:Fault");
 
         soup_message_set_status (action->msg,
                                  SOUP_STATUS_INTERNAL_SERVER_ERROR);
@@ -493,31 +503,33 @@ gupnp_service_init (GUPnPService *device)
 }
 
 /* Generate a new action response node for @action_name */
-static xmlNode *
-new_action_response_node (GUPnPService *service,
-                          const char   *action_name)
+static GString *
+new_action_response_str (GUPnPService *service,
+                         const char   *action_name)
 {
-        xmlNode *node;
+        GString *str;
         const char *type;
-        char *tmp;
 
-        tmp = g_strdup_printf ("%sResponse", action_name);
-        node = xmlNewNode (NULL, (const xmlChar *) tmp);
-        g_free (tmp);
+        str = xml_util_new_string ();
+
+        g_string_append (str, "<u:");
+        g_string_append (str, action_name);
+        g_string_append (str, "Response xmlns:u=");
 
         type = gupnp_service_info_get_service_type
                                         (GUPNP_SERVICE_INFO (service));
         if (type != NULL) {
-                xmlNs *ns;
-
-                ns = xmlNewNs (node, (xmlChar *) type, (const xmlChar *) "u");
-                xmlSetNs (node, ns);
+                g_string_append_c (str, '"');
+                g_string_append (str, type);
+                g_string_append_c (str, '"');
         } else {
                 g_warning ("No serviceType defined. Control may not work "
                            "correctly.");
         }
 
-        return node;
+        g_string_append_c (str, '>');
+
+        return str;
 }
 
 /* Handle QueryStateVariable action */
@@ -578,25 +590,26 @@ query_state_variable (GUPnPService       *service,
 
 /* controlURL handler */
 static void
-control_server_handler (SoupServerContext *server_context,
-                        SoupMessage       *msg,
+control_server_handler (SoupServer        *soup_server,
+                        SoupMessage       *msg, 
+                        const char        *server_path,
+                        GHashTable        *query,
+                        SoupClientContext *soup_client,
                         gpointer           user_data)
 {
         GUPnPService *service;
         GUPnPContext *context;
         GSSDPClient *client;
-        xmlDoc *doc, *response_doc;
-        xmlNode *action_node, *response_node;
-        xmlNs *ns;
+        xmlDoc *doc;
+        xmlNode *action_node;
         const char *soap_action, *action_name;
-        char *end, *date;
+        char *end;
+        gpointer response_body;
         GUPnPServiceAction *action;
-        xmlChar *mem;
-        int size;
 
         service = GUPNP_SERVICE (user_data);
 
-        if (strcmp (msg->method, SOUP_METHOD_POST) != 0) {
+        if (msg->method != SOUP_METHOD_POST) {
                 soup_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED);
 
                 return;
@@ -606,8 +619,8 @@ control_server_handler (SoupServerContext *server_context,
         client = GSSDP_CLIENT (context);
 
         /* Get action name */
-        soap_action = soup_message_get_header (msg->request_headers,
-                                               "SOAPAction");
+        soap_action = soup_message_headers_get (msg->request_headers,
+                                                "SOAPAction");
         action_name = strchr (soap_action, '#');
         if (!action_name) {
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
@@ -622,7 +635,8 @@ control_server_handler (SoupServerContext *server_context,
         *end = '\0';
 
         /* Parse action_node */
-        doc = xmlParseMemory (msg->request.body, msg->request.length);
+        doc = xmlParseMemory (msg->request_body->data,
+                              msg->request_body->length);
         action_node = xml_util_get_element ((xmlNode *) doc,
                                             "Envelope",
                                             "Body",
@@ -637,10 +651,10 @@ control_server_handler (SoupServerContext *server_context,
         /* Create action structure */
         action = g_slice_new (GUPnPServiceAction);
 
-        action->name          = action_name;
-        action->msg           = msg;
-        action->node          = action_node;
-        action->response_node = new_action_response_node (service, action_name);
+        action->name         = action_name;
+        action->msg          = msg;
+        action->node         = action_node;
+        action->response_str = new_action_response_str (service, action_name);
 
         /* QueryStateVariable? */
         if (strcmp (action_name, "QueryStateVariable") == 0)
@@ -661,54 +675,40 @@ control_server_handler (SoupServerContext *server_context,
                 }
         }
 
-        /* Embed action->response_node in a SOUP document */
-        response_doc = xmlNewDoc ((const xmlChar *) "1.0");
-        response_node = xmlNewDocNode (response_doc,
-                                       NULL,
-                                       (const xmlChar *) "Envelope",
-                                       NULL);
-        xmlDocSetRootElement (response_doc, response_node);
+        /* Embed action->response_str in a SOAP document */
+        g_string_prepend (action->response_str, 
+                          "<?xml version=\"1.0\"?>"
+                          "<s:Envelope xmlns:s="
+                                "\"http://schemas.xmlsoap.org/soap/envelope/\" "
+                          "s:encodingStyle="
+                                "\"http://schemas.xmlsoap.org/soap/encoding/\">"
+                          "<s:Body>");
 
-        ns = xmlNewNs (response_node,
-                       (const xmlChar *)
-                               "http://schemas.xmlsoap.org/soap/envelope/",
-                       (const xmlChar *) "s");
-        xmlSetNs (response_node, ns);
+        if (action->msg->status_code != SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+                g_string_append (action->response_str, "</u:");
+                g_string_append (action->response_str, action_name);
+                g_string_append (action->response_str, "Response>");
+        }
 
-        response_node = xmlNewChild (response_node,
-                                     ns,
-                                     (const xmlChar *) "Body",
-                                     NULL);
+        g_string_append (action->response_str,
+                         "</s:Body>"
+                         "</s:Envelope>");
 
-        if (msg->status_code == SOUP_STATUS_INTERNAL_SERVER_ERROR)
-                xmlSetNs (action->response_node, ns);
+        response_body = g_string_free (action->response_str, FALSE);
 
-        xmlAddChild (response_node, action->response_node);
-
-        /* Dump document to response */
-        xmlDocDumpMemory (response_doc, &mem, &size);
-
-        msg->response.owner  = SOUP_BUFFER_SYSTEM_OWNED;
-        msg->response.body   = g_strdup ((char *) mem);
-        msg->response.length = size;
+        soup_message_set_response (msg,
+                                   "text/xml",
+                                   SOUP_MEMORY_TAKE,
+                                   response_body,
+                                   strlen (response_body));
 
         /* Cleanup */
-        xmlFreeDoc (response_doc);
-        xmlFree (mem);
-
         g_slice_free (GUPnPServiceAction, action);
 
         /* Server header on response */
-        soup_message_add_header (msg->response_headers,
-                                 "Server",
-                                 gssdp_client_get_server_id (client));
-
-        /* Date header on response */
-        date = soup_date_generate (time (NULL));
-        soup_message_add_header (msg->response_headers,
-                                 "Date",
-                                 date);
-        g_free (date);
+        soup_message_headers_append (msg->response_headers,
+                                     "Server",
+                                     gssdp_client_get_server_id (client));
 }
 
 /* Generates a standard (re)subscription response */
@@ -726,21 +726,14 @@ subscription_response (GUPnPService *service,
         client = GSSDP_CLIENT (context);
 
         /* Server header on response */
-        soup_message_add_header (msg->response_headers,
-                                 "Server",
-                                 gssdp_client_get_server_id (client));
-
-        /* Date header on response */
-        tmp = soup_date_generate (time (NULL));
-        soup_message_add_header (msg->response_headers,
-                                 "Date",
-                                 tmp);
-        g_free (tmp);
+        soup_message_headers_append (msg->response_headers,
+                                     "Server",
+                                     gssdp_client_get_server_id (client));
 
         /* SID header */
-        soup_message_add_header (msg->response_headers,
-                                 "SID",
-                                 sid);
+        soup_message_headers_append (msg->response_headers,
+                                     "SID",
+                                     sid);
 
         /* Timeout header */
         if (timeout > 0)
@@ -748,9 +741,9 @@ subscription_response (GUPnPService *service,
         else
                 tmp = g_strdup ("infinite");
 
-        soup_message_add_header (msg->response_headers,
-                                 "Timeout",
-                                 tmp);
+        soup_message_headers_append (msg->response_headers,
+                                     "Timeout",
+                                     tmp);
         g_free (tmp);
 
         /* 200 OK */
@@ -783,10 +776,10 @@ subscription_timeout (gpointer user_data)
         return FALSE;
 }
 
-/* Parse timeout header and return its value in seconds, or -1
+/* Parse timeout header and return its value in seconds, or 0
  * if infinite */
 static int
-parse_timeout (const char *timeout)
+parse_and_limit_timeout (const char *timeout)
 {
         int timeout_seconds;
 
@@ -794,14 +787,11 @@ parse_timeout (const char *timeout)
 
         if (strncmp (timeout, "Second-", strlen ("Second-")) == 0) {
                 /* We have a finite timeout */
-                timeout_seconds = atoi (timeout + strlen ("Second-"));
-                if (timeout_seconds < GENA_MIN_TIMEOUT) {
-                        g_warning ("Specified timeout is too short. Assuming "
-                                   "default of %d.", GENA_DEFAULT_TIMEOUT);
-
-                        timeout_seconds = GENA_DEFAULT_TIMEOUT;
-                }
-        }
+                timeout_seconds = CLAMP (atoi (timeout + strlen ("Second-")),
+                                         GENA_MIN_TIMEOUT,
+                                         GENA_MAX_TIMEOUT);
+        } else
+                timeout_seconds = GENA_MAX_TIMEOUT;
 
         return timeout_seconds;
 }
@@ -817,7 +807,7 @@ subscribe (GUPnPService *service,
         char *start, *end, *uri;
         int timeout_seconds;
         GQueue *queue;
-        xmlChar *mem;
+        char *mem;
         GList *l;
 
         data = g_slice_new0 (SubscriptionData);
@@ -854,11 +844,11 @@ subscribe (GUPnPService *service,
         data->sid     = generate_sid ();
 
         /* Add timeout */
-        timeout_seconds = parse_timeout (timeout);
+        timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
-                data->timeout_id = g_timeout_add (timeout_seconds * 1000,
-                                                  subscription_timeout,
-                                                  data);
+                data->timeout_id = g_timeout_add_seconds (timeout_seconds,
+                                                          subscription_timeout,
+                                                          data);
         }
 
         /* Add to hash */
@@ -900,7 +890,7 @@ subscribe (GUPnPService *service,
         /* Cleanup */
         g_queue_free (queue);
 
-        xmlFree (mem);
+        g_free (mem);
 }
 
 /* Resubscription request */
@@ -926,11 +916,11 @@ resubscribe (GUPnPService *service,
                 data->timeout_id = 0;
         }
 
-        timeout_seconds = parse_timeout (timeout);
+        timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
-                data->timeout_id = g_timeout_add (timeout_seconds * 1000,
-                                                  subscription_timeout,
-                                                  data);
+                data->timeout_id = g_timeout_add_seconds (timeout_seconds,
+                                                          subscription_timeout,
+                                                          data);
         }
 
         /* Respond */
@@ -951,8 +941,11 @@ unsubscribe (GUPnPService *service,
 
 /* eventSubscriptionURL handler */
 static void
-subscription_server_handler (SoupServerContext *server_context,
-                             SoupMessage       *msg,
+subscription_server_handler (SoupServer        *soup_server,
+                             SoupMessage       *msg, 
+                             const char        *server_path,
+                             GHashTable        *query,
+                             SoupClientContext *soup_client,
                              gpointer           user_data)
 {
         GUPnPService *service;
@@ -960,10 +953,10 @@ subscription_server_handler (SoupServerContext *server_context,
 
         service = GUPNP_SERVICE (user_data);
 
-        callback = soup_message_get_header (msg->request_headers, "Callback");
-        nt       = soup_message_get_header (msg->request_headers, "NT");
-        timeout  = soup_message_get_header (msg->request_headers, "Timeout");
-        sid      = soup_message_get_header (msg->request_headers, "SID");
+        callback = soup_message_headers_get (msg->request_headers, "Callback");
+        nt       = soup_message_headers_get (msg->request_headers, "NT");
+        timeout  = soup_message_headers_get (msg->request_headers, "Timeout");
+        sid      = soup_message_headers_get (msg->request_headers, "SID");
 
         /* Choose appropriate handler */
         if (strcmp (msg->method, GENA_METHOD_SUBSCRIBE) == 0) {
@@ -1034,7 +1027,7 @@ gupnp_service_constructor (GType                  type,
         const GList *state_variables, *l;
         GUPnPContext *context;
         SoupServer *server;
-        SoupUri *uri;
+        SoupURI *uri;
         char *url;
 
         object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
@@ -1080,7 +1073,7 @@ gupnp_service_constructor (GType                  type,
 
         /* Get server */
         context = gupnp_service_info_get_context (info);
-        server = _gupnp_context_get_server (context);
+        server = gupnp_context_get_server (context);
 
         /* Run listener on controlURL */
         url = gupnp_service_info_get_control_url (info);
@@ -1090,8 +1083,8 @@ gupnp_service_constructor (GType                  type,
         url = soup_uri_to_string (uri, TRUE);
         soup_uri_free (uri);
 
-        soup_server_add_handler (server, url, NULL,
-                                 control_server_handler, NULL, object);
+        soup_server_add_handler (server, url,
+                                 control_server_handler, object, NULL);
         g_free (url);
 
         /* Run listener on eventSubscriptionURL */
@@ -1102,8 +1095,8 @@ gupnp_service_constructor (GType                  type,
         url = soup_uri_to_string (uri, TRUE);
         soup_uri_free (uri);
 
-        soup_server_add_handler (server, url, NULL,
-                                 subscription_server_handler, NULL, object);
+        soup_server_add_handler (server, url,
+                                 subscription_server_handler, object, NULL);
         g_free (url);
 
         return object;
@@ -1190,8 +1183,6 @@ gupnp_service_dispose (GObject *object)
 {
         GUPnPService *service;
         GObjectClass *object_class;
-        GUPnPContext *context;
-        SoupSession *session;
 
         service = GUPNP_SERVICE (object);
 
@@ -1209,24 +1200,7 @@ gupnp_service_dispose (GObject *object)
         }
 
         /* Cancel pending messages */
-        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
-        if (context)
-                session = _gupnp_context_get_session (context);
-        else
-                session = NULL; /* Not the first time dispose is called. */
-
-        while (service->priv->pending_messages) {
-                SoupMessage *msg;
-
-                msg = service->priv->pending_messages->data;
-
-                soup_message_set_status (msg, SOUP_STATUS_CANCELLED);
-                soup_session_cancel_message (session, msg);
-
-                service->priv->pending_messages =
-                        g_list_delete_link (service->priv->pending_messages,
-                                            service->priv->pending_messages);
-        }
+        g_hash_table_remove_all (service->priv->subscriptions);
 
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
@@ -1433,9 +1407,10 @@ gupnp_service_notify_valist (GUPnPService *service,
         }
 }
 
-/* Receiveid notify response. */
+/* Received notify response. */
 static void
-notify_got_response (SoupMessage *msg,
+notify_got_response (SoupSession *session,
+                     SoupMessage *msg,
                      gpointer     user_data)
 {
         SubscriptionData *data;
@@ -1447,8 +1422,7 @@ notify_got_response (SoupMessage *msg,
         data = user_data;
 
         /* Remove from pending messages list */
-        data->service->priv->pending_messages = 
-                g_list_remove (data->service->priv->pending_messages, msg);
+        data->pending_messages = g_list_remove (data->pending_messages, msg);
 
         if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
                 /* Success: reset callbacks pointer */
@@ -1456,7 +1430,7 @@ notify_got_response (SoupMessage *msg,
         } else {
                 /* Notify failed */
                 if (data->callbacks->next) {
-                        SoupUri *uri;
+                        SoupURI *uri;
                         GUPnPContext *context;
                         SoupSession *session;
 
@@ -1468,10 +1442,8 @@ notify_got_response (SoupMessage *msg,
                         soup_uri_free (uri);
 
                         /* And re-queue */
-                        data->service->priv->pending_messages = 
-                                g_list_prepend
-                                        (data->service->priv->pending_messages,
-                                         msg);
+                        data->pending_messages = 
+                                g_list_prepend (data->pending_messages, msg);
 
                         context = gupnp_service_info_get_context
                                         (GUPNP_SERVICE_INFO (data->service));
@@ -1525,27 +1497,22 @@ notify_subscriber (gpointer key,
                 return;
         }
 
-        /* Add headers */
-        soup_message_add_header (msg->request_headers,
-                                 "Content-Type",
-                                 "text/xml");
+        soup_message_headers_append (msg->request_headers,
+                                     "NT",
+                                     "upnp:event");
 
-        soup_message_add_header (msg->request_headers,
-                                 "NT",
-                                 "upnp:event");
+        soup_message_headers_append (msg->request_headers,
+                                     "NTS",
+                                     "upnp:propchange");
 
-        soup_message_add_header (msg->request_headers,
-                                 "NTS",
-                                 "upnp:propchange");
-
-        soup_message_add_header (msg->request_headers,
-                                 "SID",
-                                 data->sid);
+        soup_message_headers_append (msg->request_headers,
+                                     "SID",
+                                     data->sid);
 
         tmp = g_strdup_printf ("%d", data->seq);
-        soup_message_add_header (msg->request_headers,
-                                 "SEQ",
-                                 tmp);
+        soup_message_headers_append (msg->request_headers,
+                                     "SEQ",
+                                     tmp);
         g_free (tmp);
 
         /* Handle overflow */
@@ -1555,13 +1522,14 @@ notify_subscriber (gpointer key,
                 data->seq = 1;
 
         /* Add body */
-        msg->request.owner  = SOUP_BUFFER_SYSTEM_OWNED;
-        msg->request.body   = g_strdup (property_set);
-        msg->request.length = strlen (property_set);
+        soup_message_set_request (msg,
+                                  "text/xml",
+                                  SOUP_MEMORY_TAKE,
+                                  g_strdup (property_set),
+                                  strlen (property_set));
 
         /* Queue */
-        data->service->priv->pending_messages = 
-                g_list_prepend (data->service->priv->pending_messages, msg);
+        data->pending_messages = g_list_prepend (data->pending_messages, msg);
 
         context = gupnp_service_info_get_context
                         (GUPNP_SERVICE_INFO (data->service));
@@ -1574,68 +1542,44 @@ notify_subscriber (gpointer key,
 }
 
 /* Create a property set from @queue */
-static xmlChar *
+static char *
 create_property_set (GQueue *queue)
 {
         NotifyData *data;
-        xmlDoc *doc;
-        xmlNode *node;
-        xmlNs *ns;
-        xmlChar *mem;
-        int size;
-        char *str;
+        GString *str;
 
         /* Compose property set */
-        doc = xmlNewDoc ((const xmlChar *) "1.0");
+        str = xml_util_new_string ();
 
-        node = xmlNewDocNode (doc,
-                              NULL,
-                              (const xmlChar *) "propertyset",
-                              NULL);
-        xmlDocSetRootElement (doc, node);
-
-        ns = xmlNewNs (node,
-                       (const xmlChar *) "urn:schemas-upnp-org:event-1-0",
-                       (const xmlChar *) "e");
-        xmlSetNs (node, ns);
-
-        node = xmlNewTextChild (node,
-                                ns,
-                                (const xmlChar *) "property",
-                                NULL);
+        g_string_append (str,
+                         "<?xml version=\"1.0\"?>"
+                         "<e:propertyset xmlns:e="
+                                "\"urn:schemas-upnp-org:event-1-0\">"
+                         "<e:property>");
 
         /* Add variables */
         while ((data = g_queue_pop_head (queue))) {
-                str = gvalue_util_value_get_string (&data->value);
-                if (str != NULL) {
-                        /* Add to property set */
-                        xmlNewTextChild (node,
-                                         NULL,
-                                         (const xmlChar *) data->variable,
-                                         (xmlChar *) str);
-
-                        g_free (str);
-                }
+                xml_util_start_element (str, data->variable);
+                gvalue_util_value_append_to_xml_string (&data->value, str);
+                xml_util_end_element (str, data->variable);
 
                 /* Cleanup */
                 notify_data_free (data);
         }
 
-        /* Dump document */
-        xmlDocDumpMemory (doc, &mem, &size);
+        g_string_append (str,
+                         "</e:property>"
+                         "</e:propertyset>");
 
-        /* Cleanup */
-        xmlFreeDoc (doc);
-
-        /* Return */
-        return mem;
+        /* Cleanup & return */
+        return g_string_free (str, FALSE);
 }
 
 /* Flush all queued notifications */
 static void
 flush_notifications (GUPnPService *service)
 {
-        xmlChar *mem;
+        char *mem;
 
         /* Create property set */
         mem = create_property_set (service->priv->notify_queue);
@@ -1646,7 +1590,7 @@ flush_notifications (GUPnPService *service)
                               mem);
 
         /* Cleanup */
-        xmlFree (mem);
+        g_free (mem);
 }
 
 /**
@@ -1736,8 +1680,7 @@ strip_camel_case (char *camel_str)
                         if (i != 0 &&
                             camel_str[i + 1] != '\0' &&
                             camel_str[i - 1] != '_' &&
-                            !(g_ascii_isupper (camel_str[i - 1]) &&
-                              g_ascii_isupper (camel_str[i + 1]))) {
+                            !g_ascii_isupper (camel_str[i - 1])) {
                                 stripped[j++] = '_';
                         }
                         stripped[j++] = g_ascii_tolower (camel_str[i]);
@@ -1897,42 +1840,3 @@ gupnp_service_signals_autoconnect (GUPnPService *service,
         g_module_close (module);
         g_object_unref (introspection);
 }
-
-/**
- * _gupnp_service_new
- * @context: A #GUPnPContext
- * @root_device: The #GUPnPRootDevice
- * @element: The #xmlNode ponting to the right service element
- * @udn: The UDN of the device the service is contained in
- * @location: The location of the service description file
- * @url_base: The URL base for this service
- *
- * Return value: A #GUPnPService for the service with element @element, as
- * read from the service description file specified by @location.
- **/
-GUPnPService *
-_gupnp_service_new (GUPnPContext    *context,
-                    GUPnPRootDevice *root_device,
-                    xmlNode         *element,
-                    const char      *udn,
-                    const char      *location,
-                    const SoupUri   *url_base)
-{
-        GUPnPService *service;
-
-        g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
-        g_return_val_if_fail (element != NULL, NULL);
-        g_return_val_if_fail (location != NULL, NULL);
-        g_return_val_if_fail (url_base != NULL, NULL);
-
-        service = g_object_new (GUPNP_TYPE_SERVICE,
-                                "context", context,
-                                "root-device", root_device,
-                                "location", location,
-                                "udn", udn,
-                                "url-base", url_base,
-                                "element", element,
-                                NULL);
-
-        return service;
-}
